import os
import sys
import time
import logging

import torch
from torchvision.utils import save_image

'''
自定义一个logger类
能够生成一个目录(可以指定目录名,也可以自动根据时间来生成目录), 保存日志信息，生成图片，模型参数等
能够在终端输出日志信息，同时保存到txt文件
'''

def loadLogger(work_dir='./', save_name='log.txt', logToFile=True):
    # setup logger
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    # set prefix information
    formatter = logging.Formatter(fmt="[ %(asctime)s ] %(message)s", datefmt="%a %b %d %H:%M:%S %Y")
    sHandler = logging.StreamHandler()
    sHandler.setFormatter(formatter)
    logger.addHandler(sHandler)
    # save to file
    if logToFile:
        fHandler = logging.FileHandler(os.path.join(work_dir,save_name), mode='w')
        fHandler.setLevel(logging.DEBUG)
        fHandler.setFormatter(formatter)
        logger.addHandler(fHandler)
        # sys.stdout = fHandler.stream # 设置 print 打印到logger 日志中。
    return logger


class Logger():
    def __init__(self, base_root, runname=None, time_suffix=True, mode='train'):
        # create the dir of results
        self.base_root = base_root
        time_stamp = time.strftime("%Y%m%d%H%M%S", time.localtime())
        if runname is None:
            self.runname = time_stamp
        else:
            self.runname = f'{runname}_{time_stamp}'if time_suffix else runname
        self.res_dir=os.path.join(self.base_root,self.runname)
        if not os.path.exists(self.res_dir):
            os.makedirs(self.res_dir)

        # setup logger
        self.logger = loadLogger(self.res_dir, save_name=f'log_{mode}.txt')

    
    # log text
    def log(self,msg):
        self.logger.info(msg)

    
    # log tensor images
    def log_images(self, tensor, save_path, nrow=4):
        save_image(tensor, os.path.join(self.res_dir,save_path), nrow=nrow)


    def log_state_dict(self,state_dict,save_path):
        torch.save(state_dict, os.path.join(self.res_dir,save_path))
        self.log('save weight successfully!')


    # log tensor image
    def log_image(self,tensor,save_path):
        pass

    def log_figure(self,figure,save_path):
        pass

    def log_metric(self, name, value, global_step):
        pass

    def log_metrics(self, loss_dict):
        pass



import argparse
import re

'''
使用argparse对参数进行解析，方便通过命令行与程序进行交互
'''
def get_attrs(obj):
    attr_dict = {}
    for name in dir(obj):
        if callable(getattr(obj, name)): continue
        if re.match('__.*__', name): continue
        if re.match('__.*', name): continue
        if re.match('.*__', name): continue
        attr_dict[name] = getattr(obj, name)
    return attr_dict

def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def config_parse(config):
    attr_dict=get_attrs(config)

    parser = argparse.ArgumentParser()
    # start to parse
    for name,value in attr_dict.items():
        if isinstance(value, list):
            parser.add_argument(f'--{name}', type=type(value[0]), nargs='*', default=value)
        elif isinstance(value, bool):
            parser.add_argument(f'--{name}', type=str2bool, default=value)
        else:
            parser.add_argument(f'--{name}', type=type(value), default=value)

    args = parser.parse_args()

    # update config
    for name, value in vars(args).items():
        setattr(config, name, value)

    return config




from .base_logger import Logger

from torch.utils.tensorboard import SummaryWriter

class TensorboardLogger(Logger):
    def __init__(self, base_root, runname=None, time_suffix=True, mode='train') -> None:
        super().__init__(base_root, runname, time_suffix, mode)

        self.writer = SummaryWriter(self.res_dir)
        

    def log_metric(self, name, value, global_step):
        self.writer.add_scalar(name, value, global_step=global_step)




from .config.base_config import Config
from .config.setup import config_setup, strconfig
from .config.parse import config_parse

from .logger.base_logger import Logger
from .logger.tensorboard_logger import TensorboardLogger





import re
import json

'''
解析配置参数，接受三种参数:
1. 类对象
2. json 文件
3. dict 参数字典
返回包含参数属性的对象
'''
class Config:
    config_name = 'default'
    pass

def get_attrs(obj):
    attr_dict = {}
    for name in dir(obj):
        if callable(getattr(obj, name)): continue
        if re.match('__.*__', name): continue
        if re.match('__.*', name): continue
        if re.match('.*__', name): continue
        attr_dict[name] = getattr(obj, name)
    return attr_dict

def from_dict(dic:dict):
    config = Config()
    for name, value in dic.items():
        setattr(config, name, value)
    return config

def from_json(path):
    with open(path,'r') as f:
        param_dict = json.load(f)
        return from_dict(param_dict)

def config_setup(config):
    if isinstance(config,str):
        return from_json(config)
    elif isinstance(config,dict):
        return from_dict(config)
    else:
        return config
    
def strconfig(config):
    attr_dict = get_attrs(config)

    s = '\n# ----------------------------parameters table--------------------------- #\n'
    for name,value in attr_dict.items():
        s += f'{name}: {value}\n'
    s += '# ------------------------------------------------------------------------ #'
    return s


from .parse import config_parse
from .setup import config_setup, strconfig, get_attrs
'''
如果要从json 或者 dict 初始化参数，调用本脚的类

'''

class Config:
    '''
    初始化配置，接受三种参数:
    1. 类
    2. json 文件
    3. dict 参数字典
    '''
    def __init__(self, config, add_parse= True) -> None:
        '''
        接收参数配置，并合并
        '''
        config = config_setup(config)
        attr_dict = get_attrs(config)
        for name, value in attr_dict.items():
            setattr(self, name, value)

        if add_parse:
            self.parse()

    def parse(self):
        '''
        调用argparse，使得可以与用户进行交互
        '''
        config_parse(self)

    def __str__(self):
        '''
        打印配置信息
        '''
        return strconfig(self)


class Abstract_Logger:
    '''
    初始化时，设置保存文件夹、实验的名字、是否添加时间后缀、是否为训练模式
    logger 会在目标文件夹下，新建一个实验名字的目录，后续所有的loss，图像，模型参数都会保存在该目录下
    '''
    def __init__(self, base_root, runname=None, time_suffix=True, mode='train') -> None:
        pass


    # log text
    def log(self, msg):
        pass


    # log loss
    def log_metric(self, name, value):
        pass


    def log_state_dict(self, state_dict, save_path):
        pass


    # log tensor image
    def log_image(self, tensor, save_path):
        pass


    # log tensor images
    def log_images(self, tensor, save_path, nrow=4):
        pass


    def log_figure(self, figure, save_path):
        pass


import os, shutil
from setuptools import setup, find_packages

#移除构建的build文件夹
CUR_PATH = os.path.dirname(os.path.abspath(__file__))
path = os.path.join(CUR_PATH, 'build')
if os.path.isdir(path):
    print('INFO del dir ', path) 
    shutil.rmtree(path)


setup(
    name = 'easy_dl', #应用名
    author = 'oncemonkey',
    description="A small package for deep learning that setup config and log metric based on pytorch and tensorboard",
    version = '0.1.4',  #版本号
    url="https://github.com/OnceMonkey/easy-dl",
    packages = find_packages(),  #包括在安装包内的Python包
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)